from SPARQLWrapper import SPARQLWrapper, JSON

def query_lindas(query: str, environment: str):
    match environment:
        case "TEST":
            sparql = SPARQLWrapper("https://test.lindas.admin.ch/query")
        case "INT":
            sparql = SPARQLWrapper("https://int.lindas.admin.ch/query")
        case "PROD":
            sparql = SPARQLWrapper("https://lindas.admin.ch/query")
    sparql.setQuery(query=query)
    sparql.setReturnFormat(JSON)
    results = sparql.query().convert()
    return results["boolean"]


def cube_exists(cube_uri: str, environment: str):
    """
    This function checks whether a cube already exists in the provided environment using the Lindas query endpoint.
    If the cube already exists and the local flag is not set, the function will exit with an appropriate error message.
    Otherwise, the function will return the constructed cube URI as a URIRef object.

    Args:
        local (bool): A flag indicating whether the cube is local.
        environment (str): The environment of the cube.
    """
    query = f"ASK {{ <{cube_uri}> ?p ?o}}"
    return query_lindas(query, environment=environment)
        


import argparse
import os
import pandas as pd
import yaml
import py_cube
import logging

from py_cube.fetch import fetch
from py_cube.example import list_examples, load_example
from py_cube.cube.shared_dimension import convert_geojson_to_ttl


logger = logging.getLogger('pycube')


def serialize(input_directory: str, output_ttl: str, na_values: list[str], sep: str = ",", decimal: str = "."):
    csv_path = os.path.join(input_directory, "data.csv")
    yml_path = os.path.join(input_directory, "description.yml")
    json_path = os.path.join(input_directory, "description.json")

    if os.path.exists(yml_path):
        with open(yml_path) as file:
            cube_yaml = yaml.safe_load(file)
    elif os.path.exists(json_path):
        with open(json_path) as file:
            cube_yaml = yaml.safe_load(file)
    else:
        raise FileNotFoundError("Neither description.yml nor description.json found in the directory")

    df = pd.read_csv(csv_path, na_values=na_values, sep=sep, decimal=decimal)

    cube = py_cube.Cube(dataframe=df, cube_yaml=cube_yaml, environment="TEST", local=True)
    cube.prepare_data()
    cube.write_cube()
    cube.write_observations()
    cube.write_shape()
    cube.serialize(os.path.join(os.getcwd(), output_ttl))
    print(cube)


def configure_logging(log_level):
    class CustomFormatter(logging.Formatter):
        """Custom logging formatter to add colors based on log level."""

        COLORS = {
            'DEBUG': '\033[0m',  # Normal
            'INFO': '\033[94m',  # Blue
            'WARNING': '\033[93m',  # Yellow
            'ERROR': '\033[91m',  # Red
            'CRITICAL': '\033[91m',  # Red
        }

        def format(self, record):
            log_fmt = self.COLORS.get(record.levelname, '\033[0m') + '%(levelname)s: %(message)s\033[0m'
            formatter = logging.Formatter(log_fmt)
            return formatter.format(record)

    console_handler = logging.StreamHandler()
    logger.setLevel(log_level)
    console_handler.setFormatter(CustomFormatter())
    logger.addHandler(console_handler)


def main():
    parser = argparse.ArgumentParser(description="Cube data operations")
    subparsers = parser.add_subparsers(dest="operation", help="Operation to perform")

    serialize_parser = subparsers.add_parser("serialize", help="Serialize cube data")
    serialize_parser.add_argument("input_directory", help="Directory containing the data files")
    serialize_parser.add_argument("output_ttl", help="Output TTL file")
    serialize_parser.add_argument("--na_value", nargs="+", help="Values to treat as NA")
    serialize_parser.add_argument("--sep", default=",", nargs="?", help="Separator for CSV file")
    serialize_parser.add_argument("--decimal", default=".", nargs="?", help="Decimal separator")
    serialize_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    fetch_parser = subparsers.add_parser("fetch", help="Fetches a dataset from a URL")
    fetch_parser.add_argument("input_url", type=str, help="The URL of the dataset to fetch")
    fetch_parser.add_argument("output", type=str, help="The directory to save the output files")
    fetch_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    shared_parser = subparsers.add_parser("shared", help="Shared Dimension operations")
    shared_subparsers = shared_parser.add_subparsers(dest="suboperation", help="Shared sub-operations")

    convert_geojson_parser = shared_subparsers.add_parser("convert_geojson", help="Convert GeoJSON to TTL")
    convert_geojson_parser.add_argument("input_geojson", type=str, help="Input GeoJSON file")
    convert_geojson_parser.add_argument("output_ttl", type=str, help="Output TTL file")
    convert_geojson_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    example_parser = subparsers.add_parser("example", help="Example operations")
    example_subparsers = example_parser.add_subparsers(dest="suboperation", help="Example sub-operations")

    list_parser = example_subparsers.add_parser("list", help="List all examples")
    list_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    start_fuseki_parser = example_subparsers.add_parser("start-fuseki", help="Start a Fuseki database")
    start_fuseki_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    load_parser = example_subparsers.add_parser("load", help="Load an example by name")
    load_parser.add_argument("example_name", type=str, help="The name of the example to load", choices=[example["id"] for example in list_examples()])
    # add optional base_uri argument to load parser
    load_parser.add_argument("--base-uri", type=str, help="The base URI for a SPARQL database (Fuseki supported)", default="http://localhost:3030/dataset")
    load_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")

    schema_parser = subparsers.add_parser("schema", help="Schema operations")
    schema_subparsers = schema_parser.add_subparsers(dest="suboperation", help="Schema sub-operations")
    schema_subparsers.add_parser("import", help="Import the description schema file")
    schema_parser.add_argument("output", type=str, help="Output file")
    schema_parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase verbosity")


    args = parser.parse_args()
    log_level = logging.DEBUG if args.verbose == 1 else logging.INFO

    configure_logging(log_level)

    if args.operation == "serialize":
        serialize(args.input_directory, args.output_ttl, args.na_value, args.sep, args.decimal)
    elif args.operation == "fetch":
        fetch(args.input_url, args.output)
    elif args.operation == "example":
        if args.suboperation == "list":
            examples = list_examples()
            for example in examples:
                print(f'{example["id"]}: {example["name"]}')
        elif args.suboperation == "load":
            load_example(args.example_name, args.base_uri)
        elif args.suboperation == "start-fuseki":
            os.system("scripts/fuseki/start.sh")
    elif args.operation == "shared":
        if args.suboperation == "convert_geojson":
            convert_geojson_to_ttl(args.input_geojson, args.output_ttl)
    elif args.operation == 'schema':
        if args.suboperation == "import-description":
            description_schema_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'description.schema.json')
            with open(description_schema_path, 'r') as f:
                schema = f.read()
            with open(args.output, 'w') as f:
                f.write(schema)
                logger.debug(f"Imported description into current directory: {args.output}")



if __name__ == "__main__":
    main()


from pylindas.shared_dimension_queries.shared_dimensions_queries import list_shared_dimensions, list_shared_dimension_terms, list_shared_dimensions_print, print_sparql_result
from rdflib import URIRef

"""
Author: Fabian Cretton - HEVS

See README for an explanation
"""

def main():
    print("Shared dimensions query examples")
    print("================================")

    print("List all Shared Dimensions:")
    print("---------------------------")
    result = list_shared_dimensions("INT")
    list_shared_dimensions_print(result)

    # print("List 10 Shared Dimensions:")
    # print("--------------------------")
    # result = list_shared_dimensions("INT", "fr", 0, 10)
    # list_shared_dimensions_print(result)

    print("\nList Shared Dimensions that contains \"Canton\" in the english name")
    print("---------------------------------------------------------------")
    result = list_shared_dimensions("INT", "en", 0, 0, "Canton")
    list_shared_dimensions_print(result, "INT")

    print("\nList the Cantons shared dimension's terms")
    print("-----------------------------------------")
    result = list_shared_dimension_terms("INT", "https://ld.admin.ch/dimension/canton", "fr")
    print_sparql_result(result, ["name", "sdTerm"])

if __name__ == "__main__":
    main()



import re
from urllib.parse import quote
from rdflib import BNode, Graph, Literal, RDF, URIRef, XSD
from rdflib.collection import Collection
from datetime import datetime, timezone
try:
    from typing import Self
except:
    # fallback for Self in python 3.10
    from typing import TypeVar
    Self = TypeVar("Self", bound="SharedDimension")
import pandas as pd
import numbers
import sys

from pylindas.lindas.namespaces import *
from pylindas.lindas.query import query_lindas
from pyshacl import validate

class SharedDimension:
    _sd_uri: URIRef # Note: Shared Dimension usually do not have a trailing "/" in there URL
    _sd_dict: dict
    _graph: Graph
    _dataframe: pd.DataFrame
    _shape_dict: dict
    _shape_URI: URIRef
    _languages = ["fr", "en", "de", "it"]
    
    def __init__(self, dataframe: pd.DataFrame, sd_yaml: dict, environment: str, local=False):
        """
        Initialize a Shared Dimension (SD) object.

        Args:
            dataframe (pd.DataFrame): The Pandas DataFrame representing the shared dimension terms data.
            sd_yaml (dict): A dictionary containing SD information.
            environment (str): The environment of the SD.
            local (bool): A flag indicating whether the SD is local.

        Returns:
            None
        """
        self._dataframe = dataframe
        self._setup_sd_dict(sd_yaml=sd_yaml,local=local, environment=environment)
        assert self._sd_uri is not None
        self._graph = self._setup_graph()

    def __str__(self) -> str:
        """
        Return a string representation of the SD object.

        This method returns a string representation of the SD object, including its URI and name.

        Returns:
            str: A string representation of the SD object.
        """
        how_many_triples_query = (
            "SELECT (COUNT(*) as ?Triples)"
            "WHERE {"
            "    ?s ?p ?o."
            "}"
        )
        how_many_triples = self._graph.query(how_many_triples_query).bindings[0].get("Triples").value
        output = (f"SD Object <{self._sd_uri}> with name '{self._sd_dict.get('Name').get('en')}'.\n\n"
                  f"{self._dataframe.head()}\n"
                  f"Number of triples in Graph: {how_many_triples}")
        return output

    def prepare_data(self) -> Self:
        """
        Prepare the cube data by constructing terms URIs and applying mappings.
        Also handling links to other terms (as a parent term) if defined in the configuration.

        This method constructs terms and linked other terms URIs for each row in the dataframe and applies mappings to the dataframe.

        Returns:
            self
        """
        self._construct_terms_uri()

        terms = self._sd_dict.get("Terms")
        if "links-to-other-terms" in terms:
            self._construct_links_uri(terms.get("links-to-other-terms"))

        return self

    def write_sd(self) -> Self:
        """
        Write the SD metadata to the graph.

        Returns:
            self
        """
        self._graph.add((self._sd_uri, RDF.type, META.SharedDimension))
        self._graph.add((self._sd_uri, RDF.type, SCHEMA.DefinedTermSet))
        # self._graph.add((self._sd_uri, RDF.type, SKOS.ConceptScheme))

        self._graph.add((self._sd_uri, SCHEMA.identifier, Literal(self._sd_dict.get("Identifier"))))

        names = self._sd_dict.get("Name")
        for lan, name in names.items():
            self._graph.add((self._sd_uri, SCHEMA.name, Literal(name, lang=lan)))
            self._graph.add((self._sd_uri, DCT.title, Literal(name, lang=lan)))

        descriptions = self._sd_dict.get("Description")
        if descriptions:
            for lan, desc in descriptions.items():
                self._graph.add((self._sd_uri, SCHEMA.description, Literal(desc, lang=lan)))

        validFrom = self._sd_dict.get("Valid-from")
        if validFrom:
            self._graph.add((self._sd_uri, SCHEMA.validFrom, Literal(validFrom, datatype=XSD.dateTime)))

        contributor = self._sd_dict.get("Contributor")
        if contributor:
            contributor_node = self._write_contributor(contributor)
            self._graph.add((self._sd_uri, DCT.contributor, contributor_node))

        shacl_node = self._write_shacl()
        self._graph.add((self._sd_uri, SH.property, shacl_node))

        return self

    def get_iri(self) -> URIRef:
        return self._sd_uri

    def _setup_sd_dict(self, sd_yaml: dict, local: bool, environment="TEST") -> None:
        """
        Set up the SD uri + dictionary with the provided YAML data.

        Args:
            sd_yaml (dict): A dictionary containing SD information.

        Returns:
            None
        """
        # Note: Shared Dimension usually do not have a trailing "/" in there URL
        sd_uri = URIRef("https://ld.admin.ch/cube/dimension/" + sd_yaml.get("Identifier"))

        if not local:
            query = f"ASK {{ <{sd_uri}> ?p ?o}}"
            if query_lindas(query, environment=environment) == True:
                sys.exit("Shared Dimension already exists! Please update your yaml")

        self._sd_dict = sd_yaml
        self._sd_uri = URIRef(sd_uri)

    def _setup_graph(self) -> Graph:
        """Set up the graph by binding namespaces and returning the graph object.
        
        Returns:
            Graph: The graph object with namespaces bound.
        """
        graph = Graph()
        for prefix, nmspc in Namespaces.items():
            graph.bind(prefix=prefix, namespace=nmspc)

        return graph

    def _construct_terms_uri(self) -> None:
        """Construct terms URIs for each row in the dataframe.
        
        This function constructs terms URIs for each row in the dataframe based on the SD URI and terms identifier.
        
        Returns:
            None
        """
        termIdentifierField = self._sd_dict.get("Terms").get("identifier-field")

        # Note: Shared Dimension usually do not have a trailing "/" in there URL
        #   -> add it 
        def make_iri(row):
            return self._sd_uri + "/" + quote(row[termIdentifierField])
        self._dataframe['terms-uri'] = self._dataframe.apply(
            make_iri, axis=1
        )
        self._dataframe['terms-uri'] = self._dataframe['terms-uri'].map(URIRef)
        self._dataframe = self._dataframe.set_index("terms-uri")

    def _construct_links_uri(self, linksToOtherTerms) -> None:
        """Construct other terms URIs for each row in the dataframe.
        
        Note: the key in the dictionary is the field, in the data, that contains the other term's ID from which to build a term's URI
        IMPORTANT: this other term ID could be empty, which is the case for the root node in a hierarchy

        Returns:
            None
        """
        # Note: Shared Dimension usually do not have a trailing "/" in there URL
        #   -> add it 
        for key, value in linksToOtherTerms.items():
            keyForUri = str(key) + "-uri"

            def make_iri(row):
                if not pd.isna(row[key]):
                    return URIRef(self._sd_uri + "/" + quote(row[key]))
            
            self._dataframe[keyForUri] = self._dataframe.apply(
                make_iri, axis=1
            )

    def _write_contributor(self, contributor_dict: dict) -> BNode:
        """Writes a contributor to the graph.
        
            Args:
                contact_dict (dict): A dictionary containing information about the contributor: name and email.
                
            Returns:
                BNode or URIRef: The created BNode representing the contributor
        """

        contributor_node = BNode()
        self._graph.add((contributor_node, SCHEMA.email, Literal(contributor_dict.get("email"), datatype=XSD.string)))
        self._graph.add((contributor_node, SCHEMA.name, Literal(contributor_dict.get("name"), datatype=XSD.string)))
        return contributor_node

    def _write_shacl(self) -> BNode:
        """Writes the shacl meta-data to the graph.
            they are used by the Cube Creator when a SD is linked to a dimension of a cube and are the following triples:
			    shacl:property [ schema:name "Anhängevorrichtung"@de,
			 				"Tow coupling"@en,
			 				"Dispositif d’attelage"@fr,
			 				"Dispositivo d’aggancio rimorchio"@it ;
			 			qudt:scaleType <http://qudt.org/schema/qudt/Nominal> ] .            

            Returns:
                BNode or URIRef: The created BNode representing the shacl meta-data
        """

        shacl_node = BNode()

        names = self._sd_dict.get("Name")
        for lan, name in names.items():
            self._graph.add((shacl_node, SCHEMA.name, Literal(name, lang=lan)))

        self._graph.add((shacl_node, QUDT.scaleType, QUDT.Nominal))

        return shacl_node
    
    def write_terms(self) -> Self:
        """Write the terms of the SD

        This function iterates over the rows in the dataframe and adds each row as a term

        Returns:
            Self
        """
        terms = self._sd_dict.get("Terms")
        termIdentifierField = terms.get("identifier-field")
        termNameField= terms.get("name-field")
        multilingual= terms.get("multilingual")
        # Read the valid-from value from the configuration file, as a validFrom information will be added to each term
        validFrom = self._sd_dict.get("Valid-from")
            
        # TODO: optional-field handling: which is currently based on the handling  proposed for the "concepts" (in the respective branch)
        #   After the final decisions are taken about the "concepts" handling, apply it here
        # Prepare the optional fields of the term:
        # - the key of the entry (under the other-fields entry) being the name of the field in the data
        # - handle the URI field if it starts with a "/": adding the sd URI + the defined relative path
        if "other-fields" in terms:
            otherFields = terms.get("other-fields")

            # Note: Shared Dimension usually do not have a trailing "/" in there URL
            #   and the URI field will start with a "/" if it is a relative path
            sdUri =  str(self._sd_uri)
            
            otherFields_dict = {
                key: {
                    **value, 
                    "URI": sdUri + value["URI"] if value["URI"].startswith("/") else value["URI"]
                }
                for key, value in otherFields.items()
            }            
        else:
            otherFields_dict = {}

        if "links-to-other-terms" in terms:
            linksToOtherTerms_dict = terms.get("links-to-other-terms")
        else:
            linksToOtherTerms_dict = {}


        self._dataframe.apply(self._add_term, axis=1,  args=(termIdentifierField, termNameField, multilingual, validFrom, linksToOtherTerms_dict, otherFields_dict))
        return self

    def _add_term(self, termsData: pd.DataFrame, termIdentifierField: str, termNameField: str, multilingual: bool, validFrom, linksToOtherTerms_dict, otherFields_dict) -> None:
        """Add an observation to the cube.
        
            Args:
                termsData (pd.DataFrame): The terms data to be added
                termIdentifierField: the name of the field for the term's identifier
                termNameField: the name of the field for the term's name
                multilingual: is it a multilingual value

        
            Returns:
                None
        """
        # It seems that after _construct_terms_uri() calls _dataframe.set_index("terms-uri"), the term uri is accessible through the .name value
        self._graph.add((termsData.name, RDF.type, SCHEMA.DefinedTerm))
        self._graph.add((termsData.name, RDF.type, SD_MD.SharedDimensionTerm))

        # Add the link to the SD
        self._graph.add((termsData.name, SCHEMA.inDefinedTermSet, self._sd_uri))
        # Seems optional: add the link from the SD to the term
        self._graph.add((self._sd_uri, SCHEMA.hasDefinedTerm, termsData.name))

        self._graph.add((termsData.name, SCHEMA.identifier, Literal(termsData.get(termIdentifierField))))
        
        if multilingual:
            for lang in self._languages:
                name_key = f"{termNameField}_{lang}"
                if name_key in termsData:
                    self._graph.add((termsData.name, URIRef(SCHEMA.name), Literal(termsData.get(name_key), lang=lang)))            
        else:
            self._graph.add((termsData.name, URIRef(SCHEMA.name), Literal(termsData.get(termNameField))))

        if validFrom:
            self._graph.add((termsData.name, SCHEMA.validFrom, Literal(validFrom, datatype=XSD.dateTime)))

        # Handle links to other terms
        # The uris of the other therms were prepared in _construct_links_uri() and added to the data frame
        for key, value in linksToOtherTerms_dict.items():
            keyForUri = str(key) + "-uri"
            property = value['property']
            
            if keyForUri in termsData:
                otherTernmURL = termsData.get(keyForUri)
                if otherTernmURL: # it can be empty, for instance the root not of a hierarchy has no 'parent'
                    self._graph.add((termsData.name, URIRef(property), otherTernmURL))

        # Handling other fields/properties
        for key, value in otherFields_dict.items():
            if "multilingual" in value and value['multilingual']:
                for lang in self._languages:
                    key_lng = f"{key}_{lang}"
                    if key_lng in termsData:
                        self._graph.add((termsData.name, URIRef(value['URI']), Literal(termsData.get(key_lng), lang=lang)))
            else:
                if key in termsData:
                    rawValue = termsData.get(key)
                    if not pd.isna(rawValue):
                        # Note: get the datatype + language of the concept from the configuration file
                        # if multilingual, this is handeled here above
                        #   but it might happen that a field is not multilingual, but still concerns a specific language
                        #   on the other hand, if the 'language' key is not in the configuration file -> it will be a simple string with no language tag
                        sanitized_value = self._sanitize_value(rawValue, value.get('datatype'), value.get('language'))                        
                        self._graph.add((termsData.name, URIRef(value['URI']), sanitized_value))

    def serialize(self, filename: str) -> Self:
        """Serialize the cube to a file.

        This function serializes the cube to the given file name in turtle format.

        Args:
            filename (str): The name of the file to write the cube to.

        Returns:
            Self
        """
        self._graph.serialize(destination=filename, format="turtle", encoding="utf-8")
        return self
    
    @staticmethod
    def _sanitize_value(value, datatype, lang) -> Literal|URIRef:
        """Sanitize the input value to ensure it is in a valid format.
        
            Args:
                value: The value to be sanitized.
                datatype: The datatype of the value, given as string from XSD namespace (e.g. "integer", "string", "gYear", ...).
                lang: The language of the value if it is a string (e.g. "de", "fr", ...).
        
            Returns:
                Literal or URIRef: The sanitized value in the form of a typed or language tagged Literal or URIRef.
        """
        if pd.isna(value):
            return Literal("", datatype=CUBE.Undefined)
        elif datatype == "URI":
            return URIRef(value)
        elif lang!=None:
            return Literal(value, lang=lang)
        else:
            if datatype != None:
                return Literal(value, datatype=getattr(XSD, datatype))
            else:
                return Literal(value)

    def validate(self, shacl_file_path, result_file_path="") -> tuple[bool, str]:
        """Vaildate the shared dimension's graph with SHACL

            IMPORTANT Remark:
                This code is inspired/copied from the cube.py validate() code
                However, the SHACL file to validate a Shared Dimension is not available online currently
                We received an extract from Zazuko and saved it locally in shared_dimension_shape.ttl 
        
                This code demonstrates the validation, but should be improved when that SHACL is finalized and saved online.

            Args:
                shacl_file_path: path to the shacl file to use for validation as it is not yet available online
                    shared_dimension_shape.ttl is available for that purpose in this folder
                result_file_path: path to a turtle file to save the result of the SHACL validation. Emtpy string = result will not be saved
        """        
        valid, validation_text = self._validate_base(shacl_file_path, result_file_path)

        if valid:
            return True, "Shared dimension is valid."
        else:
            return False, validation_text

    def _validate_base(self, shacl_file_path, result_file_path=""):
        shacl_graph = Graph()
        shacl_graph.parse(shacl_file_path, format="turtle")
        valid_sd, results_graph_sd, text_sd = validate(data_graph=self._graph, shacl_graph=shacl_graph)

        if result_file_path != "":
            results_graph_sd.serialize(result_file_path, format="turtle")

        if valid_sd:
            return True, "Shared dimension basics are met."
        else:
            return False, text_sd



import json
import requests
import os

examples_dir = "example"


def load_example(example_id, base_uri="http://localhost:3030/dataset"):
    file_path = os.path.join(examples_dir, example_id, 'cube.ttl')
    with open(file_path, 'rb') as f:
        response = requests.post(base_uri, headers={"Content-Type": "text/turtle"}, data=f)
        response.raise_for_status()


def list_examples(language="en"):
    result = []
    for root, dirs, files in os.walk(examples_dir):
        if "cube.ttl" in files and "description.json" in files:
            description_path = os.path.join(root, "description.json")
            with open(description_path, 'r') as desc_file:
                desc = json.load(desc_file)
                name = desc.get("Name", {}).get(language, "")
                description = desc.get("Description", {}).get(language, "")
            result.append({
                "id": os.path.relpath(root, examples_dir),
                "name": name,
                "description": description
            })
    return result



from rdflib import Graph, Namespace


CUBE = Namespace("https://cube.link/")
DCAT = Namespace("http://www.w3.org/ns/dcat#")
DCT = Namespace("http://purl.org/dc/terms/")
FOAF = Namespace("http://xmlns.com/foaf/0.1/")
LDADMIN = Namespace("https.//ld.admin.ch/application/")
META = Namespace("https://cube.link/meta/")
QUDT = Namespace("http://qudt.org/schema/qudt/")
RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RELATION = Namespace("https://cube.link/relation/")
SCHEMA = Namespace("http://schema.org/")
SH = Namespace("http://www.w3.org/ns/shacl#")
TIME = Namespace("http://www.w3.org/2006/time#")
UNIT = Namespace("http://qudt.org/vocab/unit/")
VCARD = Namespace("http://www.w3.org/2006/vcard/ns#")
VOID = Namespace("http://rdfs.org/ns/void#")
GEO = Namespace("http://www.opengis.net/ont/geosparql#")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
SD_MD = Namespace("https://cube-creator.zazuko.com/shared-dimensions/vocab#")

Namespaces = {
    "cube": CUBE,
    "dcat": DCAT,
    "dct": DCT,
    "schema": SCHEMA,
    "sh": SH,
    "foaf": FOAF,
    "ldadmin": LDADMIN,
    "meta": META,
    "qudt": QUDT,
    "rdf": RDF,
    "relation": RELATION,
    "time": TIME,
    "unit": UNIT,
    "vcard": VCARD,
    "void": VOID,
    "geo": GEO,
    "skos": SKOS,
    "sd_md": SD_MD,
}



import pandas as pd
import sparql_dataframe
from SPARQLWrapper import SPARQLWrapper, JSON


def get_cube(endpoint: str, identifier: str, version: str):
    """Retrieve the cube URI based on the provided identifier and version using SPARQL query.
    
        Args:
            endpoint (str): The SPARQL endpoint URL.
            identifier (str): The identifier of the cube.
            version (str): The version of the cube.
    
        Returns:
            str: The URI of the cube.
    
        Raises:
            Exception: If an error occurs during the SPARQL query execution.
    """
    match endpoint:
        case "TEST":
            endpoint = "https://test.lindas.admin.ch/query"
        case "INT":
            endpoint = "https://int.lindas.admin.ch/query"
        case "PROD":
            endpoint = "https://lindas.admin.ch/query"

    query = f"""
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX cube: <https://cube.link/>
        PREFIX schema: <http://schema.org/>

        SELECT ?cube 
        {{
           ?cube a cube:Cube ;
               dcterms:identifier "{identifier}" ;
               schema:version {version} .
        }}
    """
    sparql = SPARQLWrapper(endpoint)
    sparql.setReturnFormat(JSON)
    sparql.setQuery(query)

    try:
        resp = sparql.queryAndConvert()

        return resp["results"]["bindings"][0]["cube"]["value"]
    except Exception as e:
        return e


def get_observations(endpoint: str, identifier: str, version: str):
    """Retrieve observations from a given endpoint based on the provided identifier and version.
    
        Args:
            endpoint (str): The SPARQL endpoint URL.
            identifier (str): The identifier for the observations.
            version (str): The version of the observations.
    
        Returns:
            pandas.DataFrame: A DataFrame containing the observations with columns for observation, predicate, and value.
    """
    match endpoint:
        case "TEST":
            endpoint = "https://test.lindas.admin.ch/query"
        case "INT":
            endpoint = "https://int.lindas.admin.ch/query"
        case "PROD":
            endpoint = "https://lindas.admin.ch/query"
    cube_uri = get_cube(endpoint=endpoint, identifier=identifier, version=version)
    query = f"""
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX sh: <http://www.w3.org/ns/shacl#>
        PREFIX schema: <http://schema.org/>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        PREFIX cube: <https://cube.link/>
        
        SELECT ?obs ?pred ?value
        {{
            <{cube_uri}> cube:observationSet/cube:observation ?obs .
            <{cube_uri}> cube:observationConstraint/sh:property ?dim .
            ?dim sh:path ?predURI ;
               schema:name ?pred .
            FILTER(LANG(?pred)='de')
            {{
               ?dim a cube:KeyDimension .
               ?obs ?predURI ?vl .
               ?vl schema:name ?value
            }} UNION {{
               ?dim a cube:KeyDimension .
               ?obs ?predURI ?value .
               FILTER (DATATYPE(?value) != xsd:anyURI)
            }} UNION {{
               ?dim a cube:MeasureDimension .
               ?obs ?predURI ?value
            }}
        }}
    """

    df = sparql_dataframe.get(endpoint, query)
    observations = df.pivot(index="obs", columns="pred", values="value").reset_index(drop=True)
    return observations



from pylindas.lindas.namespaces import SCHEMA
from pylindas.pycube import Cube
from rdflib import Graph
import pandas as pd
import pytest
import yaml

class TestClass:

    TEST_CASE_PATH = "example/Cubes/"

    @classmethod
    def setup_test_cube(cls, dataframe_path: str, description_path: str) -> Cube:
        with open(cls.TEST_CASE_PATH + description_path) as file:
            description = yaml.safe_load(file)
        dataframe = pd.read_csv(cls.TEST_CASE_PATH + dataframe_path)
        cube = Cube(dataframe=dataframe, cube_yaml=description, environment="TEST", local=True)
        return cube.prepare_data().write_cube(opendataswiss=True).write_observations().write_shape()

    def setup_method(self):
        self.mock_cube = self.setup_test_cube(
            "mock/data.csv", "mock/description.yml")
        self.co2_cube = self.setup_test_cube(
            "co2-limits/data.csv", "co2-limits/description.yml")
        self.hierarchies_cube = self.setup_test_cube(
            "Biotope_Statistik/biotope.csv", "Biotope_Statistik/biotope.yml")

    def test_standard_error(self):
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop schema:name 'Standardfehler'@de ;"
            "    schema:description 'Standardfehler des berechneten Werts'@de ;"
            "    sh:path mock:standardError ;"
            "    qudt:scaleType qudt:RatioScale ;"
            "    qudt:hasUnit unit:PERCENT ;"
            "    meta:dimensionRelation ["
            "      a relation:StandardError;"
            "      meta:relatesTo mock:value ;"
            "    ] ."
            "}"
        )

        result = self.mock_cube._graph.query(sparql)
        assert bool(result)

    def test_upper_uncertainty(self):
        # todo: include the co2 emission cube
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop sh:path mock:upperUncertainty ;"
            "    schema:name 'Upper Unsicherheit'@de ;"
            "    sh:maxCount 1 ;"
            "    qudt:scaleType qudt:RatioScale ;"
            "    meta:dimensionRelation ["
            "      a relation:ConfidenceUpperBound ;"
            '      dct:type "Confidence interval" ;'
            "      meta:relatesTo mock:value ;"
            "    ] ."
            "}"
        )

        # result = self.cube._graph.query(sparql)
        assert True

    def test_lower_uncertainty(self):
        # todo: include the co2 emission cube
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop schema:name 'Lower Unsicherheit'@de ;"
            "    schema:description 'Lower Unsicherheit'@de ;"
            "    sh:path mock:lowerUncertainty ;"
            "    qudt:scaleType qudt:RatioScale ;"
            "    qudt:hasUnit unit:PERCENT ;"
            "    meta:dimensionRelation ["
            "      a relation:ConfidenceLowerBound ;"
            "      dct:type 'Confidence interval' ;"
            "      meta:relatesTo mock:value ;"
            "    ] ."
            "}"
        )

        # result = self.cube._graph.query(sparql)
        assert True

    def test_point_limit(self):
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop sh:path limit_1:co2Emissions ;"
            "    meta:annotation ?annotation ."
            "  ?annotation a meta:Limit ;"
            "    schema:value 1.849298e+01 ;"
            "    meta:annotationContext ["
            "      sh:path limit_1:year ;"
            "      sh:hasValue <https://ld.admin.ch/time/year/2012> ;"
            "    ] ; "
            "    meta:annotationContext [ "
            "      sh:path limit_1:energySource ;"
            "      sh:hasValue <https://mock.ld.admin.ch/energySource/01> ;"
            "  ]."
            "}"
        )
    
        result = self.co2_cube._graph.query(sparql)
        assert bool(result)
    
    def test_range_limit(self):
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop sh:path limit_1:co2Emissions ;"
            "    meta:annotation ?annotation ."
            "  ?annotation a meta:Limit ;"
            "    schema:minValue 1.708845e+01  ;"
            "    schema:maxValue 1.779072e+01 ;"
            "    meta:annotationContext ["
            "      sh:path limit_1:year ;"
            "      sh:hasValue <https://ld.admin.ch/time/year/2016> ;"
            "    ] ; "
            "    meta:annotationContext [ "
            "      sh:path limit_1:energySource ;"
            "      sh:hasValue <https://mock.ld.admin.ch/energySource/01> ;"
            "    ] ."
            "}"
        )

        result = self.co2_cube._graph.query(sparql)
        assert bool(result)

    def test_limit_cube_validity(self):
        result_bool, result_message = self.co2_cube.validate()
        assert result_message == "Cube is valid."

    def test_annotation_dimension(self):
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop sh:path mock:status ;"
            "     schema:name 'Veröffentlichungsstatus'@de ;"
            "     qudt:scaleType qudt:NominalScale ."
            "   minus {"
            "     ?prop a cube:KeyDimension ."
            "   }"
            "   minus {"
            "     ?prop a cube:MeasureDimension ."
            "   }"
            "}"
        )

        result = self.mock_cube._graph.query(sparql)
        assert bool(result)

    def test_mock_cube_validity(self):
        result_bool, result_message = self.mock_cube.validate()
        assert result_message == "Cube is valid."

    def test_hierarchies(self):
        sparql = (
            "ASK"
            "{"
            "  ?shape a cube:Constraint ;"
            "    sh:property ?prop ."
            "  ?prop sh:path biotop:type ;"
            "    meta:inHierarchy ?hierarchy ."
            "  ?hierarchy a meta:Hierarchy ;"
            "    meta:hierarchyRoot <https://environment.ld.admin.ch/foen/biotopes/tot> ;"
            "    schema:name 'Biotope' ;"
            "    meta:nextInHierarchy ?nextInHierarchy ."
            "  ?nextInHierarchy schema:name 'Biotoparten' ;"
            "    sh:path schema:hasPart ;"
            "}"
        )

        result = self.hierarchies_cube._graph.query(sparql)
        assert bool(result)

    def test_hierarchies_cube_validity(self):
        result_bool, result_message = self.hierarchies_cube.validate()
        assert result_message == "Cube is valid."

    def test_concepts(self):
        # As setup_method() is called before each test
        # -> create the concept cube only for this specific test
        # To change later if needed
        self.concepts_cube = self.setup_test_cube("concept_table_airport/data_with_dummy.csv", "concept_table_airport/description.yml")

        # Add the concept data to the cube's data
        airport_concept_df = pd.read_csv(self.TEST_CASE_PATH + "/concept_table_airport/airportType.csv")
        self.concepts_cube.write_concept("typeOfAirport", airport_concept_df)
        # data_with_dummy.csv, the data file uploaded for that cube, contains an airport type identifier that doesn't exist in airportType.csv
        # the goal is to demonstrate that the  check_dimension_object_property() called here under will detect that
        # Check that all the generated URLs for the typeOfAirport are resources (concept) with a SCHEMA.name triple
        # This allows to check if all the entries in data_with_dummy.csv correspond to an entry in airportType.csv 
        allConceptsFound = self.concepts_cube.check_dimension_object_property("typeOfAirport", SCHEMA.name)
        # allConceptsFound should be False: the dummy airport type has no correspondance in the concepts
        assert not bool(allConceptsFound)

        # Now add the dummy airportType 
        airport_concept_dummy_df = pd.read_csv("example/Cubes/concept_table_airport/airportType_dummy.csv")
        self.concepts_cube.write_concept("typeOfAirport", airport_concept_dummy_df)
        allConceptsFound = self.concepts_cube.check_dimension_object_property("typeOfAirport", SCHEMA.name)
        # allConceptsFound should be True: the dummy airport type has been added to the concepts
        assert bool(allConceptsFound)



from rdflib import Graph, Literal, RDF, URIRef
from py_cube.lindas.namespaces import *
from shapely.geometry import shape
import json
import argparse


class GeoSharedDimension(object):
    _base_uri: URIRef
    _graph: Graph
    _description: dict

    def __init__(self, base_uri: URIRef, description: dict, graph: Graph):
        self._base_uri = base_uri
        self._graph = graph
        self._description = description
    
    def _setup_graph(self) -> Graph:
        """Set up the graph by binding namespaces and returning the graph object.
        
        Returns:
            Graph: The graph object with namespaces bound.
        """
        graph = Graph()
        for prefix, nmspc in Namespaces.items():
            graph.bind(prefix=prefix, namespace=nmspc)
        try:
            graph.bind(prefix=self._cube_dict.get("Namespace"), namespace=Namespace(self._base_uri))
        except KeyError:
            print("no Namespace")
            pass
        return graph


    def _geojson_to_wkt(self, geojson: dict) -> str:
        """Convert GeoJSON to WKT.
        
        Returns:
            str: The WKT string.
        """
        if not geojson:
            return None
        geom = shape(geojson)
        return geom.wkt


    def _add_geo_feature_to_graph(self, geojson_feature):
        properties = geojson_feature.get("properties")
        if not properties:
            raise ValueError("Feature must have properties")
        iri = properties.get("iri")
        if not iri:
            raise ValueError("Feature must have an IRI")
        feature = URIRef(iri)
        self._graph.add((feature, RDF.type, URIRef("http://schema.org/Place")))

        for lang in ["fr", "en", "de", "it"]:
            name_key = f"name_{lang}"
            if name_key in properties:
                self._graph.add((feature, URIRef("http://schema.org/name"), Literal(properties[name_key], lang=lang)))

        geometry = URIRef(f"{iri}/geometry")
        self._graph.add((feature, URIRef("http://www.opengis.net/ont/geosparql#hasGeometry"), geometry))
        wkt = self._geojson_to_wkt(geojson_feature['geometry'])
        if wkt:
            self._graph.add((geometry, URIRef("http://www.opengis.net/ont/geosparql#asWKT"), Literal(wkt, datatype=URIRef("http://www.opengis.net/ont/geosparql#wktLiteral"))))


    def serialize(self, filename: str) -> None:
        """Serialize the cube to a file.

        This function serializes the cube to the given file name in turtle format.

        Args:
            filename (str): The name of the file to write the cube to.

        Returns:
            None
        """
        self._graph.serialize(destination=filename, format="turtle", encoding="utf-8")


def convert_geojson_to_ttl(geojson_filename, ttl_filename):
    with open(geojson_filename, 'r') as f:
        geojson_data = json.load(f)

    base_uri = URIRef("http://example.org/base")
    description = {}
    graph = Graph()

    shared_dimension = GeoSharedDimension(base_uri, description, graph)

    for feature in geojson_data.get("features", []):
        print(f"Adding feature {feature['properties']['name_de']}")
        shared_dimension._add_geo_feature_to_graph(feature)

    shared_dimension.serialize(ttl_filename)



import requests
import stardog
from configparser import ConfigParser



#URL = "https://stardog-test.cluster.ldbar.ch/lindas?graph=..."
#HEADERS = {'Content-Type': 'text/turtle', 'Authorization': }

# def uplod_ttl(filename: str, named_graph: str, password: str):
#     with open(filename) as file:
#         graph = file.read()
#         response = requests.request("POST", )


def _load_config(db_file:str, environment: str) -> dict:
    parser = ConfigParser()
    parser.read(db_file)
    
    config = {}
    if parser.has_section(environment):
        params = parser.items(environment)
        for param in params:
            config[param[0]] = param[1]
    else:
        raise Exception(f"Environment '{environment}' not found in db_file")

    return config

def upload_ttl(filename: str, db_file: str, environment: str, graph_uri: str):
    conn_details = _load_config(db_file, environment)

    # todo: could graph_uri be specified in lindas.ini?
    with stardog.Connection("lindas", **conn_details) as conn:
        conn.begin()
        conn.add(stardog.content.File(file=filename), graph_uri=graph_uri)
        conn.commit()


"""
Utils to download a data.europa.eu dataset with frictionless metadata,
and generate a description.json.

TODO: Make it  more agnostic from data.europa.eu
"""

import requests
import json
import os
from datetime import datetime
from typing import Dict, Any, List
from jsonschema import Draft202012Validator, validate
from jsonschema.exceptions import ValidationError

import logging

logger = logging.getLogger('pycube')

def download_json(url):
    response = requests.get(url)
    response.raise_for_status()
    return response.json()


def read_schema(schema_path):
    with open(schema_path, 'r') as f:
        return json.load(f)

class DataEuropaFetcher(object):
    """
    In the future, the class should be splitted into frictionless parsing methods and
    data.europa.eu fetch methods.
    """
    def __init__(self):
        pass

    def _transform_url(self, input_url):
        dataset_id = input_url.split('/')[-1].split('?')[0]
        return f"https://data.europa.eu/api/hub/search/datasets/{dataset_id}"


    def _extract_metadata(self, data):
        metadata = {
            "title": {
                "en": data['result']['title'].get('en', ''),
                "de": data['result']['title'].get('de', '')
            },
            "description": {
                "en": data['result']['description'].get('en', ''),
                "de": data['result']['description'].get('de', '')
            },
            "publisher": data['result'].get('publisher', '')
        }
        return metadata


    def _get_distributions(self, distributions):
        csv_data = None
        frictionless_data = None
        for distribution in distributions:
            if distribution['title'].get('en') == "Frictionless Tabular Data Resource":
                frictionless_url = distribution['access_url'][0]
                frictionless_data = requests.get(frictionless_url).json()
            if distribution['format'].get('id') == "CSV":
                csv_url = distribution['access_url'][0]
                csv_data = requests.get(csv_url).content
        return {
            "frictionless": frictionless_data,
            "csv": csv_data
        }


    def _infer_dimension_type(self, field: Dict[Any, Any], primary_keys: List[str]) -> str:
        """Infer the dimension type based on field properties."""
        if field['name'] in primary_keys:
            return "Key Dimension"
        return "Measure Dimension"


    def _infer_scale_type(self, field: Dict[Any, Any]) -> str:
        """Infer the scale type based on field properties."""
        field_type = field.get("type")
        if field_type == "string":
            return "nominal"
        elif field_type == "integer":
            return "interval"
        elif field_type == "number":
            return "ratio"
        return "nominal"  # default


    def _infer_temporal_dimension(self, field: Dict[Any, Any]) -> bool:
        """Infer if the field is a temporal dimension."""
        field_type = field.get("type")
        if field_type == "date":
            return True
        if field_type == "time":
            return True
        field_name = field['name']
        if field_name.lower() in ["jahr", "year", "date", "datum"]:
            logger.warning(f'Dimension {field_name}: Temporal dimension inferred from field name. Please verify.')
            return True



    def _generate_dimensions(self, data_metadata: Dict[Any, Any]) -> Dict[str, Dict[Any, Any]]:
        """Generate dimensions from data metadata schema."""
        dimensions = {}
        
        primary_key = data_metadata["schema"].get('primaryKey', [])
        primary_keys = primary_key if isinstance(primary_key, list) else [primary_key]
        if not primary_keys:
            first_field = data_metadata["schema"]["fields"][0]["name"]
            logger.warning(f"Primary key not found in schema. Using first field {first_field} as primary key. You may need to adjust Key/Measure Dimension manually.")
            primary_key = first_field

        for field in data_metadata["schema"]["fields"]:
            field_name = field["name"]
            
            # Create dimension object
            dimension = {
                "name": {
                    "de": field.get("title", field_name),
                    "en": field.get("title", field_name)
                },
                "dimension-type": self._infer_dimension_type(field, primary_keys),
                "scale-type": self._infer_scale_type(field),
                "path": field_name,
                "description": {
                    "de": field.get("description", f"Beschreibung für {field_name}"),
                    "en": field.get("description", f"Description for {field_name}")
                }
            }
            
            # Add unit if present
            if "unit" in field:
                dimension["unit"] = field["unit"]
            
            # Add data-kind if temporal
            if self._infer_temporal_dimension(field):
                dimension["data-kind"] = {
                    "type": "temporal",
                    "unit": "year"
                }
                
            dimensions[field_name] = dimension
        
        return dimensions


    def _transform_metadata(self, metadata: Dict[Any, Any], data_metadata: Dict[Any, Any]) -> Dict[Any, Any]:
        """Transform metadata to conform to the JSON schema."""
        
        output = {
            "Name": {
                "de": metadata["title"]["de"],
                "en": metadata["title"]["en"]
            },
            "Description": {
                "de": metadata["description"]["de"],
                "en": metadata["description"]["en"]
            },
            "Publisher": [
                {
                    "IRI": metadata["publisher"]["resource"]
                }
            ],
            "Creator": [
                {
                    "IRI": metadata["publisher"]["resource"]
                }
            ],

            "Contributor": [],

            "Date Created": datetime.now().isoformat(),
            "Contact Point": {
                "E-Mail": "opendata@example.ch",  # Example email
                "Name": metadata["publisher"]["name"]
            },
            "Base-URI": data_metadata["path"],
            "Identifier": data_metadata["name"],
            "Version": 0.1,
            "Work Status": "Draft",
            "Visualize": True,
            "Accrual Periodicity": "",
            "Namespace": "https://opendata.example.ch",
            "dimensions": self._generate_dimensions(data_metadata)
        }
        
        return output


    def fetch_dataset(self, input_url, output_dir):
        transformed_url = self._transform_url(input_url)
        data = download_json(transformed_url)

        metadata = self._extract_metadata(data)

        data_csv_filename = os.path.join(output_dir, 'data.csv')
        description_json_filename = os.path.join(output_dir, 'description.json')
        frictionless_json_filename = os.path.join(output_dir, 'frictionless.json')

        distributions = self._get_distributions(data['result']['distributions'])

        logger.info(f"Writing {data_csv_filename}")
        with open(data_csv_filename, 'wb') as f:
            f.write(distributions['csv'])

        logger.info(f"Writing {frictionless_json_filename}")
        with open(frictionless_json_filename, 'w') as f:
            f.write(json.dumps(distributions['frictionless'], indent=2))

        current_file_dir = os.path.dirname(os.path.realpath(__file__))
        description_schema_path = os.path.join(current_file_dir, 'description.schema.json')
        description_schema = read_schema(description_schema_path)
        description = self._transform_metadata(metadata, distributions['frictionless'])
        schema_path = os.path.relpath(
            description_schema_path,
            start=os.path.join(os.getcwd(), output_dir)
        )
        description = {
            "$schema": f"{schema_path}",
            **description
        }
        logger.info(f"Writing {description_json_filename}")
        with open(description_json_filename, 'w') as f:
            f.write(json.dumps(description, indent=2))
        
        validator = Draft202012Validator(description_schema)
        errors = list(validator.iter_errors(description))
        for error in errors:
            logger.warning(f"Validation Error: {error.message}")

        serialize_command = f"""# You may want to adjust the command with --sep and --decimal, depending on the data.csv
python cli.py serialize {output_dir} {os.path.join(output_dir, 'cube.ttl')}"""
        if len(errors):
            logger.warning(f"""The data and description have been downloaded. There were validation errors during description validation, you should fix them before running the following command

{serialize_command}
""")
        else:
            logger.info(f"""Success ! The data and description have been downloaded, you may now verify it, adjust it, and then run serialize to create RDF triples

{serialize_command}
""")


def fetch(input_url: str, output_dir: str):
    # create the output dir if it does not exist
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)

    if input_url.startswith('https://data.europa.eu'):
        fetcher = DataEuropaFetcher()
        fetcher.fetch_dataset(input_url, output_dir)
    else:
        raise ValueError(f'Only supporting datasets from data.europa.eu at the moment, make sure your input URL starts with https://data.europa.eu')


from pylindas.pyshareddimension import SharedDimension
from rdflib import Graph
import pandas as pd
import pytest
import yaml

class TestClass:

    TEST_CASE_PATH = "example/Shared_Dimensions/"

    @classmethod
    def setup_test_shared_dimension(cls, dataframe_path: str, description_path: str) -> SharedDimension:
        with open(cls.TEST_CASE_PATH + description_path) as file:
            description = yaml.safe_load(file)
        dataframe = pd.read_csv(cls.TEST_CASE_PATH + dataframe_path, sep=";")
        sd = SharedDimension(dataframe=dataframe, sd_yaml=description, environment="TEST", local=True)
        return sd.prepare_data().write_sd().write_terms()

    def setup_method(self):
        self.shared_dimension = self.setup_test_shared_dimension(
            "shared_dimension_generation/sd_terms.csv", "shared_dimension_generation/sd_description.yml")

    # SHACL validation of the Shared Dimension
    #   Please see the comment of the SharedDimension.validate() method
    #   in order to understand the parameters
    # This is work in progress as the SHACL file has to be passed as parameter instead of being downloaded from the Web behind the scene
    def test_perform_SHACL_validation(self):
        result_bool, result_message = self.shared_dimension.validate("./pylindas/pyshareddimension/shared_dimension_shape.ttl")
        assert result_bool == True        

    # Test some basic triples of the shared dimension
    # Will better be done by the SHACL validation
    # However, this allows to test that triples that might be optional for SHACL are correctly generated in the example
    #   as for instance schema:validFrom
    def test_shared_dimension_triples(self):
        sparql = (
            "PREFIX meta: <https://cube.link/meta/>"
            "PREFIX schema: <http://schema.org/>"
            "PREFIX dct: <http://purl.org/dc/terms/>"
            "ASK"
            "{"
            "  ?sd a meta:SharedDimension, schema:DefinedTermSet ;"
            "     schema:name ?name ;"
            "     schema:description ?desc ;"
            "     schema:identifier ?identifier ;"
            "     dct:contributor/schema:email ?contributorEmail ;"
            "     dct:contributor/schema:name ?contributorName ;"
            "     schema:validFrom ?validFrom"
            "}"
        )

        result = self.shared_dimension._graph.query(sparql)
        assert bool(result)

    def test_shared_dimension_unwanted_triples(self):
        # A Shared dimension generated by code should not have the rdf:type:
        #   hydra:Resource, md:SharedDimension
        # Those types are given to shared dimensions generated from the Cube Creator
        # When generated by code, the shared dimension should marked 'read-only' in the Cube Creator
        #   and this is achieved by not having those 2 types
        sparql = (
            "PREFIX meta: <https://cube.link/meta/>"
            "PREFIX md: <https://cube-creator.zazuko.com/shared-dimensions/vocab#>"
            "PREFIX hydra: <https://www.w3.org/ns/hydra/core#>"
            "ASK"
            "{"
            "  ?sd a meta:SharedDimension,"
            "      hydra:Resource, md:SharedDimension"
            "}"
        )

        result = self.shared_dimension._graph.query(sparql)
        # assert that NO result is found
        assert not bool(result)        
        

    def test_shared_dimension_terms(self):
        # Find terms that are missing some triples
        sparql = (
            "PREFIX meta: <https://cube.link/meta/>"
            "PREFIX schema: <http://schema.org/>"
            "PREFIX md: <https://cube-creator.zazuko.com/shared-dimensions/vocab#>"
            "PREFIX hydra: <https://www.w3.org/ns/hydra/core#>"
            "PREFIX sd_md: <https://cube-creator.zazuko.com/shared-dimensions/vocab#>"
            "ASK"
            "{"
            "  ?sd a schema:DefinedTerm, sd_md:SharedDimensionTerm ."
            "   FILTER NOT EXISTS {?sd schema:identifier ?id;"
            "       schema:inDefinedTermSet <https://ld.admin.ch/cube/dimension/pylindas_sd_generation_example> ;"
            "       schema:name ?name ;"
            "       schema:validFrom ?validFrom ;"
            "   }"
            "}"
        )

        result = self.shared_dimension._graph.query(sparql)
        # There should be no term missing those triples
        assert not bool(result)        
        
    def test_shared_dimension_one_single_root(self):
        # In that example, there should be only one terme with no parent (skos:broader)
        sparql = (
            "PREFIX schema: <http://schema.org/>"
            "PREFIX sd_md: <https://cube-creator.zazuko.com/shared-dimensions/vocab#>"
            "PREFIX skos: <http://www.w3.org/2004/02/skos/core#>"
            "SELECT *"
            "{"
            "  ?sd a schema:DefinedTerm, sd_md:SharedDimensionTerm ."
            "   FILTER NOT EXISTS {?sd skos:broader ?parent}"
            "}"
        )

        result = self.shared_dimension._graph.query(sparql)
        # There should be only one result
        assert len(list(result)) == 1


import re
from urllib.parse import quote
from rdflib import BNode, Graph, Literal, RDF, URIRef, XSD, DCTERMS
from rdflib.collection import Collection
from datetime import datetime, timezone
try:
    from typing import Self
except:
    # fallback for Self in python 3.10
    from typing import TypeVar
    Self = TypeVar("Self", bound="Cube")
from typing import Tuple
import pandas as pd
import numbers
import sys
from pylindas.lindas.namespaces import *
from pylindas.lindas.query import query_lindas
from pyshacl import validate

class Cube:
    _base_uri: URIRef
    _cube_uri: URIRef
    _cube_uri_no_version: URIRef # same as _cube_uri but without the version, needed for concepts handling
    _cube_dict: dict
    _graph: Graph
    _dataframe: pd.DataFrame
    _shape_dict: dict
    _shape_URI: URIRef
    _languages = ["fr", "en", "de", "it"]

    
    def __init__(self, dataframe: pd.DataFrame, cube_yaml: dict, environment: str, local=False):
        """
        Initialize a Cube object.

        Args:
            dataframe (pd.DataFrame): The Pandas DataFrame representing the cube data.
            cube_yaml (dict): A dictionary containing cube information.
            environment (str): The environment of the cube.
            local (bool): A flag indicating whether the cube is local.

        Returns:
            None
        """
        self._dataframe = dataframe
        self._setup_cube_dict(cube_yaml=cube_yaml)
        self._cube_uri, self._cube_uri_no_version = self._setup_cube_uri(local=local, environment=environment)
        assert self._cube_uri is not None
        self._setup_shape_dicts()
        self._graph = self._setup_graph()
        # self._graph.serialize("example/mock-cube.ttl", format="turtle")

    def __str__(self) -> str:
        """
        Return a string representation of the Cube object.

        This method returns a string representation of the Cube object, including its URI and name.

        Returns:
            str: A string representation of the Cube object.
        """
        how_many_triples_query = (
            "SELECT (COUNT(*) as ?Triples)"
            "WHERE {"
            "    ?s ?p ?o."
            "}"
        )
        how_many_triples = self._graph.query(how_many_triples_query).bindings[0].get("Triples").value
        output = (f"Cube Object <{self._cube_uri}> with name '{self._cube_dict.get('Name').get('en')}'.\n\n"
                  f"{self._dataframe.head()}\n"
                  f"Number of triples in Graph: {how_many_triples}")
        return output

    def prepare_data(self) -> Self:
        """
        Prepare the cube data by constructing observation URIs and applying mappings.

        This method constructs observation URIs for each row in the dataframe and applies mappings to the dataframe.

        Returns:
            self
        """
        self._construct_obs_uri()
        self._apply_mappings()
        return self

    def write_cube(self, opendataswiss=False) -> Self:
        """
        Write the cube metadata to the graph.

        This method writes the cube metadata to the graph, including its URI, name, description, publisher, creator, contributor, contact point, version, and date information.

        Returns:
            self
        """
        self._graph.add((self._cube_uri, RDF.type, CUBE.Cube))
        self._graph.add((self._cube_uri, RDF.type, SCHEMA.Dataset))
        self._graph.add((self._cube_uri, RDF.type, DCAT.Dataset))
        self._graph.add((self._cube_uri, RDF.type, VOID.Dataset))

        names = self._cube_dict.get("Name")
        for lan, name in names.items():
            self._graph.add((self._cube_uri, SCHEMA.name, Literal(name, lang=lan)))

        descriptions = self._cube_dict.get("Description")
        for lan, desc in descriptions.items():
            self._graph.add((self._cube_uri, SCHEMA.description, Literal(desc, lang=lan)))

        publisher = self._cube_dict.get("Publisher")
        for pblshr in publisher:
            self._graph.add((self._cube_uri, SCHEMA.publisher, URIRef(pblshr.get("IRI"))))

        creator = self._cube_dict.get("Creator")
        for crtr in creator:
            self._graph.add((self._cube_uri, SCHEMA.creator, URIRef(crtr.get("IRI"))))

        contributor = self._cube_dict.get("Contributor")
        for cntrbtr in contributor:
            self._graph.add((self._cube_uri, SCHEMA.contributor, URIRef(cntrbtr.get("IRI"))))

        dcat_contact_point = self._write_dcat_contact_point(self._cube_dict.get("Contact Point"))
        self._graph.add((self._cube_uri, DCAT.contactPoint, dcat_contact_point))
        schema_contact_point = self._write_schema_contact_point(self._cube_dict.get("Contact Point"))
        self._graph.add((self._cube_uri, SCHEMA.contactPoint, schema_contact_point))

        for creator in self._cube_dict.get("Creator", []):
            iri = creator.get('IRI')
            self._graph.add((self._cube_uri, DCT.creator, URIRef(iri)))

        for theme in self._cube_dict.get("Themes", []):
            iri = theme.get('IRI')
            if not iri:
                continue
            self._graph.add((self._cube_uri, DCAT.theme, URIRef(theme['IRI'])))

        version = self._cube_dict.get("Version")
        self._graph.add((self._cube_uri, SCHEMA.version, Literal(version)))

        identifier = self._cube_dict.get("Identifier")
        self._graph.add((self._cube_uri, DCTERMS.identifier, Literal(identifier)))

        today = datetime.today().strftime("%Y-%m-%d")
        now = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")

        self._graph.add(
            (self._cube_uri, SCHEMA.dateCreated, Literal(self._cube_dict.get("Date Created"), datatype=XSD.date)))
        self._graph.add((self._cube_uri, SCHEMA.datePublished, Literal(today, datatype=XSD.date)))
        # todo: serialization yields improper format for datetime (with timezone as +02:00 instead of proper UTC)
        self._graph.add((self._cube_uri, SCHEMA.dateModified, Literal(now, datatype=XSD.dateTime)))

        self._graph.add((self._cube_uri, CUBE.observationSet, self._cube_uri + "/ObservationSet"))
        self._graph.add((self._cube_uri, CUBE.observationConstraint, self._shape_URI))

        if self._cube_dict.get("Visualize"):
            self._graph.add((self._cube_uri, SCHEMA.workExample, URIRef("https://ld.admin.ch/application/visualize")))

        work_status = self._cube_dict.get("Work Status")
        if work_status == "Published":
            self._graph.add((self._cube_uri, SCHEMA.creativeWorkStatus,
                             URIRef("https://ld.admin.ch/vocabulary/CreativeWorkStatus/Published")))
            if opendataswiss:
                self._add_opendata_profile()
        elif work_status == "Draft":
            self._graph.add((self._cube_uri, SCHEMA.creativeWorkStatus,
                             URIRef("https://ld.admin.ch/vocabulary/CreativeWorkStatus/Draft")))

        if self._cube_dict.get("Accrual Periodicity"):
            accrual_periodicity_uri = self._get_accrual_periodicity(self._cube_dict.get("Accrual Periodicity"))
            self._graph.add((self._cube_uri, DCT.accrualPeriodicity, accrual_periodicity_uri))
        return self

    def get_iri(self) -> URIRef:
        return self._cube_uri

    def _setup_cube_dict(self, cube_yaml: dict) -> None:
        """
        Set up the cube dictionary with the provided YAML data.

        Args:
            cube_yaml (dict): A dictionary containing cube information.

        Returns:
            None
        """
        self._base_uri = URIRef(cube_yaml.get("Base-URI"))
        self._cube_dict = cube_yaml

    def _setup_cube_uri(self, local: bool, environment="TEST") ->  Tuple[URIRef, URIRef]:
        """
        Set up the cube URI by concatenating the base URI and the cube identifier with the version.

        Returns:
            URIRef: The constructed cube URI as a URIRef object.
        """
        cube_uri_no_version = self._base_uri + "/".join(["cube", str(self._cube_dict.get("Identifier"))])
        cube_uri = cube_uri_no_version + "/" + str(self._cube_dict.get("Version"))

        query = f"ASK {{ <{cube_uri}> ?p ?o}}"
        if not local:
            if query_lindas(query, environment=environment) == True:
                sys.exit("Cube already exist! Please update your yaml")

        return URIRef(cube_uri), URIRef(cube_uri_no_version)
    
    def _setup_shape_dicts(self) -> None:
        """Set up shape dictionaries by extracting key dimensions from cube dictionary.
        
            This function initializes the shape dictionary, shape URI, and key dimensions list based on the cube dictionary.
        
            Returns:
                None
        """
        self._shape_dict = self._cube_dict.pop("dimensions")
        self._shape_URI = URIRef(self._cube_uri + "/shape") 
        self._key_dimensions = [dim_name for dim_name, dim in self._shape_dict.items() if dim.get("dimension-type") == "Key Dimension"]

    def _setup_graph(self) -> Graph:
        """Set up the graph by binding namespaces and returning the graph object.
        
        Returns:
            Graph: The graph object with namespaces bound.
        """
        graph = Graph(bind_namespaces="none")
        for prefix, nmspc in Namespaces.items():
            graph.bind(prefix=prefix, namespace=nmspc)
        try:
            graph.bind(prefix=self._cube_dict.get("Namespace"), namespace=Namespace(self._base_uri))
        except KeyError:
            print("no Namespace")
            pass
        return graph

    def _construct_obs_uri(self) -> None:
        """Construct observation URIs for each row in the dataframe.
        
        This function constructs observation URIs for each row in the dataframe based on the cube URI and key dimensions.
        
        Returns:
            None
        """
        def make_iri(row):
            return self._cube_uri + "/observation/" + "_".join([quote(str(row[key_dim])) for key_dim in self._key_dimensions])
        self._dataframe['obs-uri'] = self._dataframe.apply(
            make_iri, axis=1
        )
        self._dataframe['obs-uri'] = self._dataframe['obs-uri'].map(URIRef)
        self._dataframe = self._dataframe.set_index("obs-uri")

    # Function for dynamic replacement of column names, in between {} by effective column values in a row
    #   template example: http://the_cube_uri/concept/airport_type/{typeOfAirport}/{typeOfAirport2nd}
    def _replace_placeholders(self, row, template):
        result = template
        placeholders = re.findall(r'\{(.*?)\}', template)  # find each place holder inbetween {}
        for placeholder in placeholders:
            if placeholder in row:
                result = result.replace(f'{{{placeholder}}}', str(row[placeholder]))
        return result

    def _apply_mappings(self) -> None:
        """Apply mappings to the dataframe based on the specified mapping type.
        
        This method iterates through the dimensions in the shape dictionary and applies mappings to the dataframe if a mapping is defined for the dimension. 
        For dimensions with 'additive' mapping type, it adds a baseline URI in front of the value. For example the entry 1999 will be replaced with 
        https://ld.admin.ch/time/year/1999. 
        For dimensions with 'replace' mapping type, it replaces values in the dataframe column based on the specified replacements.
        Values are not transformed to URIRef or Literal.
        
        Returns:
            None
        """
        for dim_name, dim_dict in self._shape_dict.items():
            if "mapping" in dim_dict:
                mapping = dim_dict.get("mapping")
                match mapping.get("type"):
                    case "additive":
                        base = mapping.get("base") + "{}"
                        self._dataframe[dim_name] = self._dataframe[dim_name].map(lambda x: base.format(quote(str(x))))
                    case "replace":
                        self._dataframe[dim_name] = self._dataframe[dim_name].map(mapping.get("replacements"))
                    case "regex":
                        pat = re.compile(mapping.get("pattern"))
                        repl = mapping.get("replacement")
                        self._dataframe[dim_name] = self._dataframe[dim_name].map(lambda x: re.sub(pat, repl, x))
                    case "concept":
                        # The replacement string is a URL with fields in-between {}, as for example:
                        #   /airport_type/{typeOfAirport}/{typeOfAirport2nd}
                        repl = mapping.get("replacement-automated")
                        # If the path is relative (it starts with "/"), then happen it to the cube's URL 
                        #   It also means that the concept is generated with the cube
                        #   thus also add the hard-coded "/concept" path
                        if repl.startswith("/"):
                            # cast the URIRef to a string to avoid log warnings "does not look like a valid URI"
                            repl = str(self._cube_uri) + "/concept" + repl

                        # Perform the {} placeholder replacement with the column values, for each row
                        self._dataframe[dim_name] = self._dataframe.apply(lambda row: self._replace_placeholders(row, repl), axis=1)
                        
                value_type = mapping.get("value-type", 'Shared')
                assert value_type in ['Shared', 'Literal']
                self._dataframe[dim_name] = self._dataframe[dim_name].map(lambda v: URIRef(v) if value_type == "Shared" else Literal(v))

    def _write_dcat_contact_point(self, contact_dict: dict) -> BNode | URIRef:
        """Writes a contact point to the graph.
        
            Args:
                contact_dict (dict): A dictionary containing information about the contact point.
                
            Returns:
                BNode or URIRef: The created BNode or URIRef representing the contact point.
        """
        if contact_dict.get("IRI"):
            return URIRef(contact_dict.get("IRI"))
        else:
            contact_node = BNode()
            self._graph.add((contact_node, RDF.type, VCARD.Organization))
            self._graph.add((contact_node, VCARD.hasEmail, Literal(contact_dict.get("E-Mail"), datatype=XSD.string)))
            self._graph.add((contact_node, VCARD.fn, Literal(contact_dict.get("Name"), datatype=XSD.string)))
            return contact_node

    def _write_schema_contact_point(self, contact_dict: dict) -> BNode | URIRef:
        """Writes a contact point to the graph.

            Args:
                contact_dict (dict): A dictionary containing information about the contact point.

            Returns:
                BNode or URIRef: The created BNode or URIRef representing the contact point.
        """
        if contact_dict.get("IRI"):
            return URIRef(contact_dict.get("IRI"))
        else:
            contact_node = BNode()
            self._graph.add((contact_node, RDF.type, SCHEMA.ContactPoint))
            self._graph.add((contact_node, SCHEMA.email, Literal(contact_dict.get("E-Mail"), datatype=XSD.string)))
            self._graph.add((contact_node, SCHEMA.name, Literal(contact_dict.get("Name"), datatype=XSD.string)))
            return contact_node

    @staticmethod
    def _get_accrual_periodicity(periodicity: str) -> URIRef:
        """Get the URIRef for the given accrual periodicity.
        
        Args:
            periodicity (str): The periodicity of the accrual.
        
        Returns:
            URIRef: The URIRef corresponding to the accrual periodicity.
        """
        base_uri = URIRef("http://publications.europe.eu/resource/authority/frequency/")
        match periodicity:
            case "daily":
                return URIRef(base_uri + "DAILY")
            case "weekly":
                return URIRef(base_uri + "WEEKLY")
            case "monthly":
                return URIRef(base_uri + "MONTHLY")
            case "yearly": 
                return URIRef(base_uri + "ANNUAL")
            case "irregular":
                return URIRef(base_uri + "IRREG")

    def write_observations(self) -> Self:
        """Write observations to the cube.

        This function iterates over the rows in the dataframe and adds each row as an observation to the cube.

        Returns:
            Self
        """
        self._graph.add((self._cube_uri + "/ObservationSet", RDF.type, CUBE.ObservationSet))
        self._dataframe.apply(self._add_observation, axis=1)
        return self

    def serialize(self, filename: str) -> Self:
        """Serialize the cube to a file.

        This function serializes the cube to the given file name in turtle format.

        Args:
            filename (str): The name of the file to write the cube to.

        Returns:
            Self
        """
        self._graph.serialize(destination=filename, format="turtle", encoding="utf-8")
        return self

    def _add_observation(self, obs: pd.Series) -> None:
        """Add an observation to the cube.

        It also adds the observation URI to the observation set of the cube.
        
            Args:
                obs (pd.Series): The observation data to be added. These are the single rows from the _dataframe.
        
            Returns:
                None
        """
        self._graph.add((self._cube_uri + "/ObservationSet", CUBE.observation, obs.name)) #obs.name is the index of the row which was set to be the 'obs-uri'
        self._graph.add((obs.name, RDF.type, CUBE.Observation))
        self._graph.add((obs.name, CUBE.observedBy, URIRef(self._cube_dict.get("Creator")[0].get("IRI"))))

        for column in obs.keys():
            shape_column = self._get_shape_column(column)
            path = URIRef(self._base_uri + shape_column.get("path"))
            sanitized_value = self._sanitize_value(obs.get(column), shape_column.get("datatype"), shape_column.get("language"))
            self._graph.add((obs.name, URIRef(path), sanitized_value))

    def _get_shape_column(self, column: str):
        c = self._shape_dict.get(column)
        if not c:
            print(self._shape_dict)
            raise ValueError(f'Could not find {column}')
        return c

    def write_shape(self) -> Self:
        """Write the shape of the cube to the graph.

            This function writes the shape of the cube to the graph, which is used to validate the cube as well as for
            description of dimension metadata

            Returns:
                Self
        """
        self._graph.add((self._shape_URI, RDF.type, CUBE.Constraint))
        self._graph.add((self._shape_URI, RDF.type, SH.NodeShape))

        self._graph.add((self._shape_URI, SH.closed, Literal("true", datatype=XSD.boolean)))

        observation_class_node = self._write_observation_class_shape()
        self._graph.add((self._shape_URI, SH.property, observation_class_node))

        observed_by_node = self._write_observed_by_node()
        self._graph.add((self._shape_URI, SH.property, observed_by_node))
        for dim, dim_dict in self._shape_dict.items():
            shape = self._write_dimension_shape(dim_dict, self._dataframe[dim])
            self._graph.add((self._shape_URI, SH.property, shape))
        return self
    
    def write_concept(self, concept_key: str, concept_data: pd.DataFrame):
        """Write concepts in the cube's graph
        
        Args:
            key: The concept must be defined in the cube_yaml file, as a nested key under the "Concepts" key
            concept_data: A pandas dataframe containing values related to the concept (typically from a CSV file)
        
        Returns:
            self
        """
        if not "Concepts" in self._cube_dict:
            print(f"Error: call to write_concept() but the cube's information do not contain a \"Concepts\" key! The \"{concept_key}\" concepts will not be added to the graph.")
            return self
        
        concepts = self._cube_dict.get("Concepts")

        if not concept_key in concepts:
            print(f"Error: call to write_concept() but the cube's information do not contain a \"{concept_key}\" nested key under the \"Concepts\" key! The \"{concept_key}\" concepts will not be added to the graph.")
            return self

        concept = concepts.get(concept_key)

        # Mandatory value, will cause an exception if not found
        uri = concept.get("URI")
        nameField = concept.get("name-field")

        # Handle the URI
        # if the path is relative (it starts with "/"), then happen it to the cube's URL 
        # It also means that the concept is generated with the cube
        #   thus also add the hard-coded "/concept" path
        if uri.startswith("/"):
            # cast the URIRef to a string to avoid log warnings "does not look like a valid URI"
            uri_versioned = str(self._cube_uri) + "/concept" + uri
            # If the concept is part of the cube (URI based on the cube's URI), its URL contains the cube's version
            # Then link that version to a URI with no version as done by the Cube Creator to keep links between versions of a concept
            uri_unversioned = str(self._cube_uri_no_version) + "/concept" + uri
            concept_data["URI_UNVERSIONED"] = concept_data.apply(lambda row: self._replace_placeholders(row, uri_unversioned), axis=1)
        else:
            uri_versioned = uri


        concept_data["URI"] = concept_data.apply(lambda row: self._replace_placeholders(row, uri_versioned), axis=1)

        # Optional values
        if "multilingual" in concept:
            multilingual = concept.get("multilingual")
        else:
            multilingual = False

        if "position-field" in concept:
            positionField = concept.get("position-field")
        else:
            positionField = ""

        # Prepare the optional fields of the concept by:
        # - taking only the fields that are founc in the data
        #   remark: the key of the entry (under the other-fields entry) being the name of the field in the data
        # - handle the URI field if it starts with a "/": adding the cubeURI + "/concept/prop" + the defined relative path
        # If we wanted to take only the fields that are found in the data, we could add:
        #    for key, value in otherFields.items() if key in concept_data
        #  But this does not work now that multilingual fields are handled (the key will be suffixed with the langage tag, in the data)
        if "other-fields" in concept:
            otherFields = concept.get("other-fields")
            cubeUri =  str(self._cube_uri)
            
            otherFields_dict = {
                key: {
                    **value, 
                    "URI": cubeUri + "/concept/prop" + value["URI"] if value["URI"].startswith("/") else value["URI"]
                }
                for key, value in otherFields.items()
            }            
        else:
            otherFields_dict = {}

        # Add the concepts to the graph
        concept_data.apply(self._add_concept, axis=1,  args=(nameField, multilingual, positionField, otherFields_dict))

        return self
    
    def _add_concept(self, concept: pd.DataFrame, nameField: str, multilingual: bool, positionField:str, otherFields_dict):
        """Add a concept to the graph.
        
            Args:
                concept (pd.DataFrame): The concept data to be added (original file/csv values)
                nameField: name of the field containing the name of the concept
                multilingual: if true, a multilingual concept is created, looking for columns named 'nameField' + '_' + a pre-defined language
                positionField: name of the field that contains a position for the concept. Empty if no position to be added
                otherFields_dict: an optional dictionary value that contains other fields to add to a concept (or an empty dict if no other fields)
                    About that dictionary:
                        - the key is the name of the field in the data
                        - URI field: the URI of the property, must be prepared before hand if it is a relative path in the yaml file 
                        - multilingual:  optional, look for the fields named 'key' + '_' + a pre-defined language 

            Note: from observing concepts in Lindas, a schema:sameAs is added to a generic URL based on the cube's URL but without the version

            Returns:
                None
        """
        conceptURI = URIRef(concept.URI)
        if multilingual:
            for lang in self._languages:
                name_key = f"{nameField}_{lang}"
                if name_key in concept:
                    self._graph.add((conceptURI, URIRef(SCHEMA.name), Literal(concept.get(name_key), lang=lang)))            
        else:
            self._graph.add((conceptURI, URIRef(SCHEMA.name), Literal(concept.get(nameField))))

        if positionField:
            self._graph.add((conceptURI, URIRef(SCHEMA.position), Literal(concept.get(positionField))))

        # If the concept is part of the cube (URI based on the cube's URI), its URL contains the cube's version
        # Then link that version to a URI with no version as done by the Cube Creator to keep links between versions of a concept
        if "URI_UNVERSIONED" in concept:
            self._graph.add((conceptURI, URIRef(SCHEMA.sameAs), URIRef(concept.URI_UNVERSIONED)))

        # Handling other fields/properties
        for key, value in otherFields_dict.items():
            if "multilingual" in value and value['multilingual']:
                for lang in self._languages:
                    key_lng = f"{key}_{lang}"
                    if key_lng in concept:
                        self._graph.add((conceptURI, URIRef(value['URI'] ), Literal(concept.get(key_lng), lang=lang)))
            else:
                if key in concept:
                    # Note: get the datatype + language of the concept from the configuration file
                    # if multilingual, this is handeled here above
                    #   but it might happen that a field is not multilingual, but still concerns a specific language
                    #   on the other hand, if the 'language' key is not in the configuration file -> it will be a simple string with no language tag
                    sanitized_value = self._sanitize_value(concept.get(key), value.get('datatype'), value.get('language'))
                    self._graph.add((conceptURI, URIRef(value['URI'] ), sanitized_value))



        
    def check_dimension_object_property(self, dimension_name: str, property: URIRef) -> bool:
        try:
            dimension = self._get_shape_column(dimension_name) # raises an exception if dimension not found
            path = URIRef(self._base_uri + dimension.get("path"))

            # Prepare the SPARQL query
            query = f"""
            SELECT DISTINCT ?obj ?value
            WHERE {{
                ?obs <{path}> ?obj .
                FILTER NOT EXISTS {{?obj <{property}> ?value}}
            }}
            """

            # Execute the query
            results = self._graph.query(query)

            print("\nChecking links to concept table")
            print("--------------------------------") 
            print("Checking that objects of the '" + dimension_name + "' dimension link to a concept with a '" + str(property) + "' property")
      
            # Print the objects that have no match with that property
            allValuesFound = True
            if results: 
                allValuesFound = False
                print("Result: problem with the following concept(s):")
                for row in results:
                    print("- ", row["obj"])
            else:
                print("Result: no problem found")
                
            print("--------------------------------") 

            return allValuesFound
        except Exception as e:
            print(f"An error occurred: {e}")
            return False

    def _write_dimension_shape(self, dim_dict: dict, values: pd.Series) -> BNode:
        """Write dimension shape based on the provided dictionary and values.
        
        Args:
            dim_dict (dict): A dictionary containing information about the dimension.
            values (pd.Series): A pandas Series containing values related to the dimension.
        
        Returns:
            BNode: The created dimension node in the graph.
        """
        dim_node = BNode()
        
        self._graph.add((dim_node, SH.minCount, Literal(1)))
        self._graph.add((dim_node, SH.maxCount, Literal(1)))

        for lan, name in dim_dict.get("name").items():
            self._graph.add((dim_node, SCHEMA.name, Literal(name, lang=lan)))
        for lan, desc in dim_dict.get("description").items():
            self._graph.add((dim_node, SCHEMA.description, Literal(desc, lang=lan)))
        
        self._graph.add((dim_node, SH.path, URIRef(self._base_uri + dim_dict.get("path"))))

        match dim_dict.get("dimension-type"):
            case "Key Dimension":
                self._graph.add((dim_node, RDF.type, CUBE.KeyDimension))
                self._graph.add((dim_node, SH.nodeKind, SH.IRI))
                self._graph.add((dim_node, SH.nodeKind, SH.IRI))
                
            case "Measure Dimension":
                self._graph.add((dim_node, RDF.type, CUBE.MeasureDimension))
                self._graph.add((dim_node, SH.nodeKind, SH.Literal))
            
            case "Annotation":
                self._graph.add((dim_node, SH.nodeKind, SH.Literal))

            case "Standard Error":
                relation_node = BNode()
                relation_path = dim_dict.get("relates-to")
                self._graph.add((relation_node, RDF.type, RELATION.StandardError))
                self._graph.add((relation_node, META.relatesTo, URIRef(self._base_uri + relation_path)))
                self._graph.add((dim_node, META.dimensionRelation, relation_node))
                self._graph.add((dim_node, SH.nodeKind, SH.Literal))
            
            # todo: in Doc beschreiben
            case "Upper uncertainty":
                relation_node = BNode()
                relation_path = dim_dict.get("relates-to")
                self._graph.add((relation_node, RDF.type, RELATION.ConfidenceUpperBound))
                self._graph.add((relation_node, META.relatesTo, URIRef(self._base_uri + relation_path)))
                self._graph.add((dim_node, META.dimensionRelation, relation_node))
                self._graph.add((dim_node, SH.nodeKind, SH.Literal))
                self._graph.add((relation_node, DCT.type, Literal("Confidence interval")))

            case "Lower uncertainty":
                relation_node = BNode()
                relation_path = dim_dict.get("relates-to")
                self._graph.add((relation_node, RDF.type, RELATION.ConfidenceLowerBound))
                self._graph.add((relation_node, META.relatesTo, URIRef(self._base_uri + relation_path)))
                self._graph.add((dim_node, META.dimensionRelation, relation_node))
                self._graph.add((dim_node, SH.nodeKind, SH.Literal))
                self._graph.add((relation_node, DCT.type, Literal("Confidence interval")))

            case _ as unrecognized:
                print(f"Dimension Type '{unrecognized}' is not recognized")
        
        match dim_dict.get("scale-type"):
            case "nominal":
                self._graph.add((dim_node, QUDT.scaleType, QUDT.NominalScale))
                if dim_dict.get("dimension-type") == "Key Dimension":
                    self._add_sh_list(dim_node, values)
            case "ordinal":
                self._graph.add((dim_node, QUDT.scaleType, QUDT.OrdinalScale))
                if dim_dict.get("dimension-type") == "Key Dimension":
                    self._add_sh_list(dim_node, values)
            case "interval":
                self._graph.add((dim_node, QUDT.scaleType, QUDT.IntervalScale))
                self._add_min_max(dim_node, values)
            case "ratio":
                self._graph.add((dim_node, QUDT.scaleType, QUDT.RatioScale))
                self._add_min_max(dim_node, values)
            case _ as unrecognized:
                print(f"Scale Type '{unrecognized}' is not recognized")
        
        
        # unit from https://www.qudt.org/doc/DOC_VOCAB-UNITS.html

        if dim_dict.get("unit") is not None:
            self._graph.add((dim_node, QUDT.hasUnit, getattr(UNIT, dim_dict.get("unit"))))

        try:
            data_kind = dim_dict.get("data-kind")
            try: 
                match data_kind.get("type"):
                    case "temporal":
                        data_kind_node = BNode()
                        self._graph.add((data_kind_node, RDF.type, TIME.GeneralDateTimeDescription))
                        self._graph.add((data_kind_node, TIME.unitType, TIME.unitYear))
                        self._graph.add((dim_node, META.dataKind, data_kind_node))
                    case "spatial-shape":
                        data_kind_node = BNode()
                        self._graph.add((data_kind_node, RDF.type, SCHEMA.GeoShape))
                        self._graph.add((dim_node, META.dataKind, data_kind_node))
                    case "spatial-coordinates":
                        data_kind_node = BNode()
                        self._graph.add((data_kind_node, RDF.type, SCHEMA.GeoCoordinates))
                        self._graph.add((dim_node, META.dataKind, data_kind_node))
            except AttributeError:
                pass
        except KeyError or AttributeError:
            pass
        
        if dim_dict.get("annotation"):
            for antn in dim_dict.get("annotation"):
                annotation_node = self._write_annotation(antn)
                self._graph.add((dim_node, META.annotation, annotation_node))

        if dim_dict.get("hierarchy"):
            for hrch in dim_dict.get("hierarchy"):
                hierarchy_node = self._write_hierarchy(hrch, dim_dict)
                self._graph.add((dim_node, META.inHierarchy, hierarchy_node))

        return dim_node

    def _write_observation_class_shape(self):
        observation_class_shape = BNode()
        self._graph.add((observation_class_shape, SH.path, RDF.type))
        self._graph.add((observation_class_shape, SH.nodeKind, SH.IRI))

        list_node = BNode()
        Collection(self._graph, list_node, [CUBE.Observation])
        self._graph.add((observation_class_shape, URIRef(SH + "in"), list_node))
        return observation_class_shape

    def _write_observed_by_node(self):
        observed_by_node = BNode()
        self._graph.add((observed_by_node, SH.path, CUBE.observedBy))
        self._graph.add((observed_by_node, SH.nodeKind, SH.IRI))

        list_node = BNode()
        Collection(self._graph, list_node, [URIRef(self._cube_dict.get("Creator")[0].get("IRI"))])
        self._graph.add((observed_by_node, URIRef(SH + "in"), list_node))
        return observed_by_node

    def _write_hierarchy(self, hierarchy_dict:dict, dim_dict = None) -> BNode:
        hierarchy_node = BNode()
        self._graph.add((hierarchy_node, RDF.type, META.Hierarchy))

        root = str(hierarchy_dict.get("root"))
        if root.startswith("http"):
            self._graph.add((hierarchy_node, META.hierarchyRoot, URIRef(root)))
        else:
            mapping_dict = dim_dict.get("mapping")
            if mapping_dict.get("type") == "additive":
                root_uri = mapping_dict.get("base") + root
                self._graph.add((hierarchy_node, META.hierarchyRoot, URIRef(root_uri)))
            elif mapping_dict.get("type") == "replace":
                root_uri = mapping_dict.get("replacements").get(root)
                self._graph.add((hierarchy_node, META.hierarchyRoot, URIRef(root_uri)))

        name = hierarchy_dict.get("name")
        self._graph.add((hierarchy_node, SCHEMA.name, Literal(name)))

        next_dict = hierarchy_dict.get("next-in-hierarchy")
        self._write_next_in_hierarchy(next_dict, parent_node=hierarchy_node)
        
        return hierarchy_node

    def _write_next_in_hierarchy(self, next_dict: dict, parent_node: BNode):
        next_node = BNode()
        self._graph.add((next_node, SCHEMA.name, Literal(next_dict.get("name"))))
        self._graph.add((next_node, SH.path, URIRef(next_dict.get("path"))))

        self._graph.add((parent_node, META.nextInHierarchy, next_node))

        if next_dict.get("next-in-hierarchy"):
            self._write_next_in_hierarchy(next_dict.get("next-in-hierarchy"), next_node)

    def _write_annotation(self, annotation_dict: dict) -> BNode:
        annotation_node = BNode()
        for lan, name in annotation_dict.get("name").items():
            self._graph.add((annotation_node, SCHEMA.name, Literal(name, lang=lan)))

        if annotation_dict.get("context"):
            for dimension, value in annotation_dict.get("context").items():
                dimension_dict = self._shape_dict.get(dimension)
                dimension_path = dimension_dict.get("path")

                type_of_mapping = dimension_dict.get("mapping").get("type")
                match type_of_mapping:
                    case "additive":
                        value = dimension_dict.get("mapping").get("base") + str(value)
                    case "replace":
                        value = dimension_dict.get("mapping").get("replacements").get(value)

                context_node = BNode()
                self._graph.add((context_node, SH.path, URIRef(self._base_uri + dimension_path)))
                self._graph.add((context_node, SH.hasValue, URIRef(value)))

                self._graph.add((annotation_node, META.annotationContext, context_node))

        match annotation_dict.get("type"):
            case "limit":
                self._graph.add((annotation_node, RDF.type, META.Limit))
                self._graph.add((annotation_node, SCHEMA.value, Literal(annotation_dict.get("value"))))
            case "limit-range":
                self._graph.add((annotation_node, RDF.type, META.Limit))
                self._graph.add((annotation_node, SCHEMA.minValue, Literal(annotation_dict.get("min-value"))))
                self._graph.add((annotation_node, SCHEMA.maxValue, Literal(annotation_dict.get("max-value"))))

        return annotation_node
    
    def _add_sh_list(self, dim_node: BNode, values: pd.Series):
        """Add a SHACL list of all unique values to the given dimension node.
        
            Args:
                dim_node (BNode): The dimension node to which the SHACL list will be added.
                values (pd.Series): The values to be added to the SHACL list.
        
            Returns:
                None
        """
        list_node = BNode()
        unique_values = values.unique()
        Collection(self._graph, list_node, [URIRef(vl) for vl in unique_values])
        self._graph.add((dim_node, URIRef(SH + "in"), list_node))

    def _add_min_max(self, dim_node: BNode, values: pd.Series):
        """Add minimum and maximum values to the given dimension node.
        
            Args:
                dim_node (BNode): The dimension node to which the values will be added.
                values (pd.Series): The series of values from which minimum and maximum will be calculated.
        
            Todo:
                Case of cube.Undefined should be covered.
        """
        # todo: case of cube.Undefined should be covered
        _min = values.min()
        _max = values.max()
        self._graph.add((dim_node, SH.min, Literal(_min)))
        self._graph.add((dim_node, SH.max, Literal(_max)))

    @staticmethod
    def _sanitize_value(value, datatype, lang) -> Literal|URIRef:
        """Sanitize the input value to ensure it is in a valid format.
        
            Args:
                value: The value to be sanitized.
                datatype: The datatype of the value, given as string from XSD namespace (e.g. "integer", "string", "gYear", ...).
                lang: The language of the value if it is a string (e.g. "de", "fr", ...).
        
            Returns:
                Literal or URIRef: The sanitized value in the form of a typed or language tagged Literal or URIRef.
        """
        if pd.isna(value):
            return Literal("", datatype=CUBE.Undefined)
        elif datatype == "URI":
            return URIRef(value)
        elif lang!=None:
            return Literal(value, lang=lang)
        else:
            if datatype != None:
                return Literal(value, datatype=getattr(XSD, datatype))
            else:
                return Literal(value)

    def validate(self) -> tuple[bool, str]:
        valid, validation_text = self._validate_base()

        if self._cube_dict.get("Visualize"):
            valid_visualize, visualization_text = self._validate_visualize_profile()
            validation_text += "\n" + visualization_text
            valid = valid and valid_visualize

        if self._cube_dict.get("Opendataswiss"):
            valid_opendataswiss, opendataswiss_text = self._validate_opendata_profile()
            validation_text += "\n" + opendataswiss_text
            valid = valid and valid_opendataswiss

        if valid:
            return True, "Cube is valid."
        else:
            return False, validation_text

    def _validate_base(self, serialize_results=False):
        # first step: standalone-cube-constraint
        # Remark: standalone-cube-constraint contains standalone-constraint-constraint!
        shacl_graph = Graph()
        shacl_graph.parse("https://cube.link/latest/shape/standalone-cube-constraint", format="turtle")
        valid_cube, results_graph_cube, text_cube = validate(data_graph=self._graph, shacl_graph=shacl_graph)

        # third step: self-consistency
        # to do this, add the cube:Observations as target of the cube:Constraint
        self._graph.add((self._shape_URI, SH.targetClass, CUBE.Observation))
        consistent, results_graph_consistency, text_consistency = validate(data_graph=self._graph)
        # remove the target again
        self._graph.remove((self._shape_URI, SH.targetClass, CUBE.Observation))

        if serialize_results:
            results_graph = results_graph_cube + results_graph_consistency
            results_graph.serialize("./validation_results.ttl", format="turtle")

        if valid_cube and consistent:
            return True, "Cube basics are met."
        else:
            result_text = f"{text_cube}\n{text_consistency}"
            return False, result_text

    def _validate_visualize_profile(self, serialize_results=False):
        shacl_graph = Graph()
        shacl_graph.parse("https://cube.link/latest/shape/profile-visualize", format="turtle")

        valid, results_graph, text = validate(data_graph=self._graph, shacl_graph=shacl_graph)

        if serialize_results:
            results_graph.serialize("./validation_results.ttl", format="turtle")
        return valid, text

    def _add_opendata_profile(self):
        names = self._cube_dict.get("Name")
        for lan, name in names.items():
            self._graph.add((self._cube_uri, SCHEMA.name, Literal(name, lang=lan)))

        descriptions = self._cube_dict.get("Description")
        for lan, desc in descriptions.items():
            self._graph.add((self._cube_uri, DCT.description, Literal(desc, lang=lan)))

        self._graph.add((self._cube_uri, SCHEMA.workExample, URIRef("https://ld.admin.ch/application/opendataswiss")))

    def _validate_opendata_profile(self, serialize_results=False):
        shacl_graph = Graph()
        shacl_graph.parse("https://cube.link/latest/shape/profile-opendataswiss-lindas", format="turtle")

        valid, results_graph, text = validate(data_graph=self._graph, shacl_graph=shacl_graph)

        if serialize_results:
            results_graph.serialize("./validation_results.ttl", format="turtle")
        return valid, text



from pylindas.pycube import Cube
from pylindas.getter.get import get_cube, get_observations
from pylindas.lindas.namespaces import Namespaces
from pylindas.lindas.upload import upload_ttl
from pylindas.lindas.query import query_lindas

__version__ = "0.4.16"



from SPARQLWrapper import SPARQLWrapper, JSON
from rdflib import URIRef
from typing import List
import json

"""
Author: Fabian Cretton - HEVS

The goal of this file is to become a tool for developers to find a useful shared dimension,
then get the URLs of the terms in order to configure the mapping for a cube's dimension.

It is not yet a class with methods, and contains code that could be more generic.
For instance, query_lindas could be a very generic function as the one found in /lindas/query.py
But existing query_lindas() is specific for ASK queries (returns a bool value)

See an example usage in example_sd.py

This is a first implementation of:
- Basic queries to request shared dimensions information from LINDAS
- Display the results, line by line
"""

def query_lindas(query: str, environment: str):
    """
    Send a SPARQL query to a LINDAS end-point and return the JSON result
    Note: the values of the different environments URL should come from a config file/environment variables 
    """
    match environment:
        case "PROD":
            endpoint = "https://lindas.admin.ch/query"
        case "INT":
            endpoint = "https://int.lindas.admin.ch/query"
        case _:
            endpoint = "https://test.lindas.admin.ch/query"

    sparql = SPARQLWrapper(endpoint)
    sparql.setQuery(query=query)
    sparql.setReturnFormat(JSON)
    return sparql.query().convert()

def list_shared_dimensions(environment: str, name_lng: str="en", offset: int=0, limit: int=0, search_word: str=""):
    """
    List existing Shared Dimensions in a specific environment 
        Returns the JSON object of the SPARQL query result

    Args:
        limit: no limit if 0

    If a SD has a validThrough date, it could be deprecated (depending on the current date)
    """
    query = f"""
        PREFIX meta: <https://cube.link/meta/>
        PREFIX schema: <http://schema.org/>
        SELECT * WHERE {{
            ?sd a meta:SharedDimension .
            OPTIONAL{{ ?sd schema:name ?name .}}
            FILTER(lang(?name) = \"{name_lng}\") 
            OPTIONAL{{?sd schema:validFrom ?validFrom}}
            OPTIONAL{{?sd schema:validThrough ?validThrough}}
        """

    if search_word != "":
        query += f"FILTER contains(?name,\"{search_word}\")"

    query += f"""
        }} 
        ORDER BY ?name 
        OFFSET {offset}
        """
    if limit != 0:
        query += f"LIMIT {limit}"
    
    #print(query)
    return query_lindas(query, environment=environment)

def list_shared_dimensions_print(result: json, environment_for_terms: str=""):
    """
    Print the result of the list_shared_dimensions() query
        To the console, in a friendly manner, one sd per line with its URL, label, validFrom and validThrough values

    Args:
        environment_for_terms: if an environment is passed, for each shared dimension 2 terms will be queried and displayed
            This possibility to display 2 terms by querying LINDAS is just a POC, should be better refined
    """
    # Pretty print the JSON - for debuging purpose
    #print(json.dumps(result, indent=4))
    
    # Loop through the "bindings" and display dimensions name and URL (sd)
    if 'results' in result and 'bindings' in result['results'] and result['results']['bindings']:
        for item in result['results']['bindings']:
            # Extract the 'sd' and 'name' values
            sd = item['sd']['value']
            
            if 'name' in item:
                name = item['name']['value']
            else:
                name = "(no name in that language)"
            
            if 'validFrom' in item:
                validFrom = "- validFrom " + item['validFrom']['value']
            else:
                validFrom = ""

            if 'validThrough' in item:
                validThrough = "- validThrough " + item['validThrough']['value']
            else:
                validThrough = ""

            print(f"{name} <{sd}> {validFrom} {validThrough}")

            # if <parameter to define> -> list 2 terms for that sd
            if environment_for_terms != "":
                termsResult = list_shared_dimension_terms(environment_for_terms, sd, "en", 0, 2)
                print("{ Terms sample:")
                print_sparql_result(termsResult, ["name", "sdTerm"])
                print("}")

    else:
        print("No result binding found in that JSON result") 

def list_shared_dimension_terms(environment: str, sd_URL: URIRef, name_lng: str="en", offset: int=0, limit: int=0):
    """
    List the terms URL of a Shared Dimensions in a specific environment 
        Returns the JSON object of the SPARQL query result

    Args:
        limit: no limit if 0
    """
    query = f"""
        PREFIX schema: <http://schema.org/>
        SELECT * WHERE {{
            ?sdTerm  schema:inDefinedTermSet <{sd_URL}> .
            OPTIONAL{{?sdTerm schema:name ?name .}}
            FILTER(lang(?name) = \"{name_lng}\") 
        }} 
        ORDER BY ?name 
        OFFSET {offset}
        """
    
    if limit != 0:
        query += f"LIMIT {limit}"
    
    #print(query)
    return  query_lindas(query, environment=environment)

def print_sparql_result(result: json, fields: List[str]):
    """
    Print line by line the result of a sparql query, according to the fields in the list parameter
    - Each field is tested for existance (this function do not know about the mandatory/OPTIONAL field in the original query)
    - If a value starts with "http" -> it is displayed inbetween <>
    """

    if 'results' in result and 'bindings' in result['results'] and result['results']['bindings']:
        for item in result['results']['bindings']:
            line = ""
            for field in fields:
                if field in item:
                    fieldValue = item[field]['value']
                    if fieldValue.lower().startswith("http"):
                        fieldValue = "<" + fieldValue + ">"
                else:
                    fieldValue = ""
            
                line += fieldValue + " "

            print(line)
    else:
        print("No result binding found in that JSON result") 


from pyshacl import validate
from rdflib import Graph

def validate(data_graph: Graph, shacl_graph: Graph) -> tuple:
    conforms, results_graph, text = validate(data_graph, shacl_graph=shacl_graph, abort_on_first=True, inference="none", advanced=True)
    return conforms, text